import torch
import torch.nn as nn
import math
import torch.nn.functional as F

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class Embeddings(nn.Module):
    """
    Implements embeddings of the words and adds their positional encodings. 
    """
    def __init__(self, vocab_size, d_model, max_len = 50):
        super(Embeddings, self).__init__()
        self.d_model = d_model
        self.dropout = nn.Dropout(0.1)
        self.embed = nn.Embedding(vocab_size, d_model)
        self.pe = self.create_positinal_encoding(max_len, self.d_model)
        self.dropout = nn.Dropout(0.1)
        
    def create_positinal_encoding(self, max_len, d_model):
        pe = torch.zeros(max_len, d_model).to(device)
        for pos in range(max_len):   # for each position of the word
            for i in range(0, d_model, 2):   # for each dimension of the each position
                pe[pos, i] = math.sin(pos / (10000 ** ((2 * i)/d_model)))
                pe[pos, i + 1] = math.cos(pos / (10000 ** ((2 * (i + 1))/d_model)))
        pe = pe.unsqueeze(0)   # include the batch size
        return pe
        
    def forward(self, encoded_words):
        embedding = self.embed(encoded_words) * math.sqrt(self.d_model)
        embedding += self.pe[:, :embedding.size(1)]   # pe will automatically be expanded with the same batch size as encoded_words
        embedding = self.dropout(embedding)
        return embedding



class MultiHeadAttention(nn.Module):
    
    def __init__(self, heads, d_model):
        
        super(MultiHeadAttention, self).__init__()
        assert d_model % heads == 0
        self.d_k = d_model // heads
        self.heads = heads
        self.dropout = nn.Dropout(0.1)
        self.query = nn.Linear(d_model, d_model)
        self.key = nn.Linear(d_model, d_model)
        self.value = nn.Linear(d_model, d_model)
        self.concat = nn.Linear(d_model, d_model)
        
    def forward(self, query, key, value, mask):
        """
        query, key, value of shape: (batch_size, max_len, 512)
        mask of shape: (batch_size, 1, 1, max_words)
        """
        # (batch_size, max_len, 512)
        query = self.query(query)
        key = self.key(key)        
        value = self.value(value)   
        
        # (batch_size, max_len, 512) --> (batch_size, max_len, h, d_k) --> (batch_size, h, max_len, d_k)
        query = query.view(query.shape[0], -1, self.heads, self.d_k).permute(0, 2, 1, 3)   
        key = key.view(key.shape[0], -1, self.heads, self.d_k).permute(0, 2, 1, 3)  
        value = value.view(value.shape[0], -1, self.heads, self.d_k).permute(0, 2, 1, 3)  
        
        # (batch_size, h, max_len, d_k) matmul (batch_size, h, d_k, max_len) --> (batch_size, h, max_len, max_len)
        scores = torch.matmul(query, key.permute(0,1,3,2)) / math.sqrt(query.size(-1))
        scores = scores.masked_fill(mask == 0, -1e9)    # (batch_size, h, max_len, max_len)
        weights = F.softmax(scores, dim = -1)           # (batch_size, h, max_len, max_len)
        weights = self.dropout(weights)
        # (batch_size, h, max_len, max_len) matmul (batch_size, h, max_len, d_k) --> (batch_size, h, max_len, d_k)
        context = torch.matmul(weights, value)
        # (batch_size, h, max_len, d_k) --> (batch_size, max_len, h, d_k) --> (batch_size, max_len, h * d_k)
        context = context.permute(0,2,1,3).contiguous().view(context.shape[0], -1, self.heads * self.d_k)
        # (batch_size, max_len, h * d_k)
        interacted = self.concat(context)
        return interacted 



class FeedForward(nn.Module):

    def __init__(self, d_model, middle_dim = 2048):
        super(FeedForward, self).__init__()
        
        self.fc1 = nn.Linear(d_model, middle_dim)
        self.fc2 = nn.Linear(middle_dim, d_model)
        self.dropout = nn.Dropout(0.1)

    def forward(self, x):
        out = F.relu(self.fc1(x))
        out = self.fc2(self.dropout(out))
        return out


class EncoderLayer(nn.Module):

    def __init__(self, d_model, heads):
        super(EncoderLayer, self).__init__()
        self.layernorm = nn.LayerNorm(d_model)
        self.self_multihead = MultiHeadAttention(heads, d_model)
        self.feed_forward = FeedForward(d_model)
        self.dropout = nn.Dropout(0.1)

    def forward(self, embeddings, mask):
        interacted = self.dropout(self.self_multihead(embeddings, embeddings, embeddings, mask))
        interacted = self.layernorm(interacted + embeddings)
        feed_forward_out = self.dropout(self.feed_forward(interacted))
        encoded = self.layernorm(feed_forward_out + interacted)
        return encoded


class DecoderLayer(nn.Module):
    
    def __init__(self, d_model, heads):
        super(DecoderLayer, self).__init__()
        self.layernorm = nn.LayerNorm(d_model)
        self.self_multihead = MultiHeadAttention(heads, d_model)
        self.src_multihead = MultiHeadAttention(heads, d_model)
        self.feed_forward = FeedForward(d_model)
        self.dropout = nn.Dropout(0.1)
        
    def forward(self, embeddings, encoded, src_mask, target_mask):
        query = self.dropout(self.self_multihead(embeddings, embeddings, embeddings, target_mask))
        query = self.layernorm(query + embeddings)
        interacted = self.dropout(self.src_multihead(query, encoded, encoded, src_mask))
        interacted = self.layernorm(interacted + query)
        feed_forward_out = self.dropout(self.feed_forward(interacted))
        decoded = self.layernorm(feed_forward_out + interacted)
        return decoded


class Transformer(nn.Module):
    
    def __init__(self, d_model, heads, num_layers, word_map):
        super(Transformer, self).__init__()
        
        self.d_model = d_model
        self.vocab_size = len(word_map)
        self.embed = Embeddings(self.vocab_size, d_model)
        self.encoder = nn.ModuleList([EncoderLayer(d_model, heads) for _ in range(num_layers)])
        self.decoder = nn.ModuleList([DecoderLayer(d_model, heads) for _ in range(num_layers)])
        self.logit = nn.Linear(d_model, self.vocab_size)
        
    def encode(self, src_words, src_mask):
        src_embeddings = self.embed(src_words)
        for layer in self.encoder:
            src_embeddings = layer(src_embeddings, src_mask)
        return src_embeddings
    
    def decode(self, target_words, target_mask, src_embeddings, src_mask):
        tgt_embeddings = self.embed(target_words)
        for layer in self.decoder:
            tgt_embeddings = layer(tgt_embeddings, src_embeddings, src_mask, target_mask)
        return tgt_embeddings
        
    def forward(self, src_words, src_mask, target_words, target_mask):
        encoded = self.encode(src_words, src_mask)
        decoded = self.decode(target_words, target_mask, encoded, src_mask)
        out = F.log_softmax(self.logit(decoded), dim = 2)
        return out



# -*- coding: utf-8 -*-


from collections import Counter
import json
import torch
import torch.nn as nn

from torch.utils.data import Dataset, DataLoader
import torch.utils.data
import math
import torch.nn.functional as F
import csv

from transformers import BertTokenizer

tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')

device = torch.device("auto" if torch.cuda.is_available() else "cpu")



def remove_punc(string):
    punctuations = '''!()-[]{};:'"\,<>./?@#$%^&*_~'''
    no_punct = ""
    for char in string:
        if char not in punctuations:
            no_punct = no_punct + char  # space is also a character
    return no_punct.lower()


def tokenize_text(text):
    tokens = tokenizer.tokenize(text)
    return tokens





"""pairs = []

with open(dataset_name, 'r', errors='ignore') as csv_file:
    csv_reader = csv.reader(csv_file)

    header = next(csv_reader)

    input_index = header.index(input_set)
    label_index = header.index(label_set)
    for row in csv_reader:
        qa_pairs = []
        try:
            input_text = row[input_index]
            label_text = row[label_index]
            first = input_text.strip()
            second = label_text.strip()
            qa_pairs.append(encode_question(first))
            qa_pairs.append(encode_reply(second))
            pairs.append(qa_pairs)
        except IndexError:
            print(f"Skipping row: {row}")
            continue

with open('pairs_encoded.json', 'w') as p:
    json.dump(pairs, p)
"""

# rev_word_map = {v: k for k, v in word_map.items()}
# ' '.join([rev_word_map[v] for v in pairs_encoded[1][0]])

"""class Dataset(Dataset):

    def __init__(self):
        self.pairs = pairs
        self.dataset_size = len(self.pairs)

    def __getitem__(self, i):
        question = torch.LongTensor(self.pairs[i][0])
        reply = torch.LongTensor(self.pairs[i][1])

        return question, reply

    def __len__(self):
        return self.dataset_size"""


"""train_loader = torch.utils.data.DataLoader(Dataset(),
                                           batch_size=batch_size,
                                           shuffle=True,
                                           pin_memory=True)"""


# question, reply = next(iter(train_loader))

def create_masks(question, reply_input, reply_target,device):
    def subsequent_mask(size):
        mask = torch.triu(torch.ones(size, size)).transpose(0, 1).type(dtype=torch.uint8)
        return mask.unsqueeze(0)

    question_mask = question != 0
    question_mask = question_mask.to(device)
    question_mask = question_mask.unsqueeze(1).unsqueeze(1)  # (batch_size, 1, 1, max_words)

    reply_input_mask = reply_input != 0
    reply_input_mask = reply_input_mask.unsqueeze(1)  # (batch_size, 1, max_words)
    reply_input_mask = reply_input_mask & subsequent_mask(reply_input.size(-1)).type_as(reply_input_mask.data)
    reply_input_mask = reply_input_mask.unsqueeze(1)  # (batch_size, 1, max_words, max_words)
    reply_target_mask = reply_target != 0  # (batch_size, max_words)

    return question_mask, reply_input_mask, reply_target_mask


class Embeddings(nn.Module):
    """
    Implements embeddings of the words and adds their positional encodings.
    """

    def __init__(self, vocab_size, d_model, max_len=160):
        super(Embeddings, self).__init__()
        self.d_model = d_model
        self.dropout = nn.Dropout(0.1)
        self.embed = nn.Embedding(vocab_size, d_model)
        self.pe = self.create_positinal_encoding(max_len, self.d_model)
        self.dropout = nn.Dropout(0.1)

    def create_positinal_encoding(self, max_len, d_model):
        pe = torch.zeros(max_len, d_model).to(device)
        for pos in range(max_len):  # for each position of the word
            for i in range(0, d_model, 2):  # for each dimension of the each position
                pe[pos, i] = math.sin(pos / (10000 ** ((2 * i) / d_model)))
                pe[pos, i + 1] = math.cos(pos / (10000 ** ((2 * (i + 1)) / d_model)))
        pe = pe.unsqueeze(0)  # include the batch size
        return pe

    def forward(self, encoded_words):
        embedding = self.embed(encoded_words) * math.sqrt(self.d_model)
        embedding += self.pe[:,
                     :embedding.size(1)]  # pe will automatically be expanded with the same batch size as encoded_words
        embedding = self.dropout(embedding)
        return embedding


class MultiHeadAttention(nn.Module):

    def __init__(self, heads, d_model):
        super(MultiHeadAttention, self).__init__()
        assert d_model % heads == 0
        self.d_k = d_model // heads
        self.heads = heads
        self.dropout = nn.Dropout(0.1)
        self.query = nn.Linear(d_model, d_model)
        self.key = nn.Linear(d_model, d_model)
        self.value = nn.Linear(d_model, d_model)
        self.concat = nn.Linear(d_model, d_model)

    def forward(self, query, key, value, mask):
        """
        query, key, value of shape: (batch_size, max_len, 512)
        mask of shape: (batch_size, 1, 1, max_words)
        """
        # (batch_size, max_len, 512)
        query = self.query(query)
        key = self.key(key)
        value = self.value(value)

        # (batch_size, max_len, 512) --> (batch_size, max_len, h, d_k) --> (batch_size, h, max_len, d_k)
        query = query.view(query.shape[0], -1, self.heads, self.d_k).permute(0, 2, 1, 3)
        key = key.view(key.shape[0], -1, self.heads, self.d_k).permute(0, 2, 1, 3)
        value = value.view(value.shape[0], -1, self.heads, self.d_k).permute(0, 2, 1, 3)

        # (batch_size, h, max_len, d_k) matmul (batch_size, h, d_k, max_len) --> (batch_size, h, max_len, max_len)
        scores = torch.matmul(query, key.permute(0, 1, 3, 2)) / math.sqrt(query.size(-1))
        scores = scores.masked_fill(mask == 0, -1e9)  # (batch_size, h, max_len, max_len)
        weights = F.softmax(scores, dim=-1)  # (batch_size, h, max_len, max_len)
        weights = self.dropout(weights)
        # (batch_size, h, max_len, max_len) matmul (batch_size, h, max_len, d_k) --> (batch_size, h, max_len, d_k)
        context = torch.matmul(weights, value)
        # (batch_size, h, max_len, d_k) --> (batch_size, max_len, h, d_k) --> (batch_size, max_len, h * d_k)
        context = context.permute(0, 2, 1, 3).contiguous().view(context.shape[0], -1, self.heads * self.d_k)
        # (batch_size, max_len, h * d_k)
        interacted = self.concat(context)
        return interacted


class FeedForward(nn.Module):

    def __init__(self, d_model, middle_dim=2048):
        super(FeedForward, self).__init__()

        self.fc1 = nn.Linear(d_model, middle_dim)
        self.fc2 = nn.Linear(middle_dim, d_model)
        self.dropout = nn.Dropout(0.1)

    def forward(self, x):
        out = F.relu(self.fc1(x))
        out = self.fc2(self.dropout(out))
        return out


class EncoderLayer(nn.Module):

    def __init__(self, d_model, heads):
        super(EncoderLayer, self).__init__()
        self.layernorm = nn.LayerNorm(d_model)
        self.self_multihead = MultiHeadAttention(heads, d_model)
        self.feed_forward = FeedForward(d_model)
        self.dropout = nn.Dropout(0.1)

    def forward(self, embeddings, mask):
        interacted = self.dropout(self.self_multihead(embeddings, embeddings, embeddings, mask))
        interacted = self.layernorm(interacted + embeddings)
        feed_forward_out = self.dropout(self.feed_forward(interacted))
        encoded = self.layernorm(feed_forward_out + interacted)
        return encoded


class DecoderLayer(nn.Module):

    def __init__(self, d_model, heads):
        super(DecoderLayer, self).__init__()
        self.layernorm = nn.LayerNorm(d_model)
        self.self_multihead = MultiHeadAttention(heads, d_model)
        self.src_multihead = MultiHeadAttention(heads, d_model)
        self.feed_forward = FeedForward(d_model)
        self.dropout = nn.Dropout(0.1)

    def forward(self, embeddings, encoded, src_mask, target_mask):
        query = self.dropout(self.self_multihead(embeddings, embeddings, embeddings, target_mask))
        query = self.layernorm(query + embeddings)
        interacted = self.dropout(self.src_multihead(query, encoded, encoded, src_mask))
        interacted = self.layernorm(interacted + query)
        feed_forward_out = self.dropout(self.feed_forward(interacted))
        decoded = self.layernorm(feed_forward_out + interacted)
        return decoded


class Transformer(nn.Module):

    def __init__(self, d_model, heads, num_layers, word_map):
        super(Transformer, self).__init__()

        self.d_model = d_model
        self.vocab_size = len(word_map)
        self.embed = Embeddings(self.vocab_size, d_model)
        self.encoder = nn.ModuleList([EncoderLayer(d_model, heads) for _ in range(num_layers)])
        self.decoder = nn.ModuleList([DecoderLayer(d_model, heads) for _ in range(num_layers)])
        self.logit = nn.Linear(d_model, self.vocab_size)

    def encode(self, src_words, src_mask):
        src_embeddings = self.embed(src_words)
        for layer in self.encoder:
            src_embeddings = layer(src_embeddings, src_mask)
        return src_embeddings

    def decode(self, target_words, target_mask, src_embeddings, src_mask):
        tgt_embeddings = self.embed(target_words)
        for layer in self.decoder:
            tgt_embeddings = layer(tgt_embeddings, src_embeddings, src_mask, target_mask)
        return tgt_embeddings

    def forward(self, src_words, src_mask, target_words, target_mask):
        encoded = self.encode(src_words, src_mask)
        decoded = self.decode(target_words, target_mask, encoded, src_mask)
        out = F.log_softmax(self.logit(decoded), dim=2)
        return out


class AdamWarmup:

    def __init__(self, model_size, warmup_steps, optimizer):
        self.model_size = model_size
        self.warmup_steps = warmup_steps
        self.optimizer = optimizer
        self.current_step = 0
        self.lr = 0

    def get_lr(self):
        return self.model_size ** (-0.5) * min(self.current_step ** (-0.5),
                                               self.current_step * self.warmup_steps ** (-1.5))

    def step(self):
        # Increment the number of steps each time we call the step function
        self.current_step += 1
        lr = self.get_lr()
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr
        # update the learning rate
        self.lr = lr
        self.optimizer.step()

    def state_dict(self):
        return self.optimizer.state_dict()

    def load_state_dict(self, state_dict):
        self.optimizer.load_state_dict(state_dict)


class LossWithLS(nn.Module):

    def __init__(self, size, smooth):
        super(LossWithLS, self).__init__()
        self.criterion = nn.KLDivLoss(size_average=False, reduce=False)
        self.confidence = 1.0 - smooth
        self.smooth = smooth
        self.size = size

    def forward(self, prediction, target, mask):
        """
        prediction of shape: (batch_size, max_words, vocab_size)
        target and mask of shape: (batch_size, max_words)
        """
        prediction = prediction.view(-1, prediction.size(-1))  # (batch_size * max_words, vocab_size)
        target = target.contiguous().view(-1)  # (batch_size * max_words)
        mask = mask.float()
        mask = mask.view(-1)  # (batch_size * max_words)
        labels = prediction.data.clone()
        labels.fill_(self.smooth / (self.size - 1))
        labels.scatter_(1, target.data.unsqueeze(1), self.confidence)
        loss = self.criterion(prediction, labels)  # (batch_size * max_words, vocab_size)
        loss = (loss.sum(1) * mask).sum() / mask.sum()
        return loss


def evaluate(transformer, question, question_mask, max_len):
    transformer.eval()
    start_token = tokenizer.cls_token_id
    encoded = transformer.encode(question, question_mask)
    words = torch.LongTensor([[start_token]]).to(device)

    for step in range(max_len - 1):
        size = words.shape[1]
        target_mask = torch.triu(torch.ones(size, size)).transpose(0, 1).type(dtype=torch.uint8)
        target_mask = target_mask.to(device).unsqueeze(0).unsqueeze(0)
        decoded = transformer.decode(words, target_mask, encoded, question_mask)
        predictions = transformer.logit(decoded[:, -1])
        _, next_word = torch.max(predictions, dim=1)
        next_word = next_word.item()
        if next_word == tokenizer.sep_token_id:
            break
        words = torch.cat([words, torch.LongTensor([[next_word]]).to(device)], dim=1)

    # Convert tensor to list of token ids
    words = words.squeeze(0).tolist()

    # Convert token ids to tokens
    tokens = tokenizer.convert_ids_to_tokens(words)

    # Remove special tokens and join the tokens into a sentence
    sentence = ' '.join(tokens).replace(tokenizer.pad_token, '').strip()

    return sentence


"""transformer = Transformer(d_model = d_model, heads = heads, num_layers = num_layers, word_map = tokenizer.vocab)
transformer = transformer.to(device)
adam_optimizer = torch.optim.Adam(transformer.parameters(), lr=0, betas=(0.9, 0.98), eps=1e-9)
transformer_optimizer = AdamWarmup(model_size = d_model, warmup_steps = 2000, optimizer = adam_optimizer)
criterion = LossWithLS(len(tokenizer.vocab), 0.1)
total_params = sum(p.numel() for p in transformer.parameters())

print(f"Total number of parameters: {total_params}")


def train(train_loader, transformer, criterion, epoch):
    transformer.train()
    sum_loss = 0
    count = 0

    for i, (question, reply) in enumerate(train_loader):
        samples = question.shape[0]

        # Move to device
        question = question.to(device)
        reply = reply.to(device)

        # Prepare Target Data
        reply_input = reply[:, :-1]
        reply_target = reply[:, 1:]

        # Create mask and add dimensions
        question_mask, reply_input_mask, reply_target_mask = create_masks(question, reply_input, reply_target)

        # Get the transformer outputs
        out = transformer(question, question_mask, reply_input, reply_input_mask)

        # Compute the loss
        loss = criterion(out, reply_target, reply_target_mask)

        # Backprop
        transformer_optimizer.optimizer.zero_grad()
        loss.backward()
        transformer_optimizer.step()

        sum_loss += loss.item() * samples
        count += samples

        if i % 100 == 0:
            print("Epoch [{}][{}/{}]\tLoss: {:.3f}".format(epoch + 1, i, len(train_loader), sum_loss / count))


for epoch in range(epochs):

    train(train_loader, transformer, criterion, epoch)

    state = {'epoch': epoch+1, 'transformer': transformer, 'transformer_optimizer': transformer_optimizer,'train_state': transformer.state_dict(),'optimizer_state_dict': transformer_optimizer.state_dict()}
    torch.save(state, model_name+'_' + str(epoch+1) + '.pth.tar')



model_weights_path = ""
state = {'epoch': epochs, 'transformer': transformer, 'transformer_optimizer': transformer_optimizer,'train_state': transformer.state_dict(),'optimizer_state_dict': transformer_optimizer.state_dict()}

torch.save(transformer.state_dict(), model_weights_path)

checkpoint = torch.load(model_name)
transformer = checkpoint['transformer']

def inference(transformer,max_len,question):


    enc_qus = tokenizer.encode(question, add_special_tokens=True, max_length=max_len, truncation=True)
    question = torch.LongTensor(enc_qus).to(device).unsqueeze(0)
    question_mask = (question != tokenizer.pad_token_id).to(device).unsqueeze(1).unsqueeze(1)
    sentence = evaluate(transformer, question, question_mask, max_len)
    print(sentence)

# Implementation of tokenizer till here
"""


class CustomDataset(Dataset):
    def __init__(self, pairs):
        self.pairs = pairs
        self.dataset_size = len(self.pairs)

    def __getitem__(self, i):
        question = torch.LongTensor(self.pairs[i][0])
        reply = torch.LongTensor(self.pairs[i][1])
        return question, reply

    def __len__(self):
        return self.dataset_size


class TransformerModel:
    def __init__(self, d_model, dataset_name, input_set, label_set, batch_size, heads, num_layers, device, epochs,
                 model_name, max_len=512, transformer=None):
        self.tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
        self.max_len = max_len
        self.dataset_name = dataset_name
        self.input_set = input_set
        self.label_set = label_set
        self.batch_size = batch_size
        self.d_model = d_model
        self.heads = heads
        self.num_layers = num_layers
        self.device = device
        self.epochs = epochs
        self.model_name = model_name
        self.pairs = self.load_and_process_data()
        self.train_loader = self.create_data_loader()
        if transformer is None:
            self.transformer = self.create_transformer()
        else:
            self.transformer = transformer
        self.criterion = LossWithLS(len(self.tokenizer.vocab), 0.1)
        self.optimizer = self.create_optimizer()

    def load_and_process_data(self):
        pairs = []
        with open(self.dataset_name, 'r', errors='ignore') as csv_file:
            csv_reader = csv.reader(csv_file)
            header = next(csv_reader)
            input_index = header.index(self.input_set)
            label_index = header.index(self.label_set)
            for row in csv_reader:
                try:
                    input_text = row[input_index].strip()
                    label_text = row[label_index].strip()
                    qa_pairs = [self.encode_question(input_text), self.encode_reply(label_text)]
                    pairs.append(qa_pairs)
                except IndexError:
                    print(f"Skipping row: {row}")
                    continue
        return pairs

    def create_data_loader(self):
        dataset = CustomDataset(self.pairs)
        return DataLoader(dataset, batch_size=self.batch_size, shuffle=True, pin_memory=True)

    def create_transformer(self):
        transformer = Transformer(self.d_model, self.heads, self.num_layers, self.tokenizer.vocab)
        total_params = sum(p.numel() for p in transformer.parameters())
        print(f"Total parameters of model: {total_params}")
        return transformer.to(self.device)

    def create_optimizer(self):
        adam_optimizer = torch.optim.Adam(self.transformer.parameters(), lr=0, betas=(0.9, 0.98), eps=1e-9)
        return AdamWarmup(model_size=self.d_model, warmup_steps=2000, optimizer=adam_optimizer)

    def encode_question(self, text):
        tokens = self.tokenizer.tokenize(text)
        token_ids = self.tokenizer.convert_tokens_to_ids(tokens)
        token_ids += [self.tokenizer.pad_token_id] * (self.max_len - len(token_ids))
        return token_ids

    def encode_reply(self, text):
        tokens = self.tokenizer.tokenize(text)
        token_ids = [self.tokenizer.cls_token_id] + self.tokenizer.convert_tokens_to_ids(tokens) + [
            self.tokenizer.sep_token_id]
        token_ids += [self.tokenizer.pad_token_id] * (self.max_len - len(token_ids))
        return token_ids

    def train(self,save_checkpoint=True):
        for epoch in range(self.epochs):
            self.transformer.train()
            sum_loss = 0
            count = 0
            for i, (question, reply) in enumerate(self.train_loader):
                samples = question.shape[0]
                question = question.to(self.device)
                reply = reply.to(self.device)
                reply_input = reply[:, :-1]
                reply_target = reply[:, 1:]
                question_mask, reply_input_mask, reply_target_mask = create_masks(question, reply_input, reply_target,
                                                                                  self.device)
                out = self.transformer(question, question_mask, reply_input, reply_input_mask)
                loss = self.criterion(out, reply_target, reply_target_mask)
                self.optimizer.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
                sum_loss += loss.item() * samples
                count += samples
                if i % 100 == 0:
                    print(f"Epoch [{epoch + 1}][{i}/{len(self.train_loader)}]\tLoss: {sum_loss / count:.3f}")
            if save_checkpoint == True:
                state = {
                    'epoch': epoch + 1,
                    'transformer': self.transformer,
                    'transformer_optimizer': self.optimizer,
                    'train_state': self.transformer.state_dict(),
                    'optimizer_state_dict': self.optimizer.state_dict()
                }
                torch.save(state, f"{self.model_name}_{epoch + 1}.pth.tar")
            else:
                pass


        state = {
            'epoch': self.epochs,
            'transformer': self.transformer,
            'transformer_optimizer': self.optimizer,
            'train_state': self.transformer.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict()
        }
        torch.save(state, f"{self.model_name}_final.pth.tar")

    def inference(self, question):
        self.transformer.eval()
        enc_qus = tokenizer.encode(question, add_special_tokens=True, max_length=self.max_len, truncation=True)
        question = torch.LongTensor(enc_qus).to(device).unsqueeze(0)
        question_mask = (question != tokenizer.pad_token_id).to(device).unsqueeze(1).unsqueeze(1)
        sentence = evaluate(self.transformer, question, question_mask, self.max_len)
        return sentence




import json
import torch
from torch.utils.data import Dataset
import torch.utils.data



train_loader = torch.utils.data.DataLoader(Dataset(),
                                           batch_size = 64,
                                           shuffle=True,
                                           pin_memory=True)

d_model = 512
heads = 8
num_layers = 3
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
epochs = 100

with open('WORDMAP_corpus.json', 'r') as j:
    word_map = json.load(j)
    
transformer = Transformer(d_model = d_model, heads = heads, num_layers = num_layers, word_map = word_map)
transformer = transformer.to(device)
adam_optimizer = torch.optim.Adam(transformer.parameters(), lr=0, betas=(0.9, 0.98), eps=1e-9)
transformer_optimizer = AdamWarmup(model_size = d_model, warmup_steps = 4000, optimizer = adam_optimizer)
criterion = LossWithLS(len(word_map), 0.1)
total_params = sum(p.numel() for p in transformer.parameters())
print(f"Total number of parameters: {total_params}")


def train(train_loader, transformer, criterion, epoch):
    
    transformer.train()
    sum_loss = 0
    count = 0

    for i, (question, reply) in enumerate(train_loader):
        
        samples = question.shape[0]

        # Move to device
        question = question.to(device)
        reply = reply.to(device)

        # Prepare Target Data
        reply_input = reply[:, :-1]
        reply_target = reply[:, 1:]

        # Create mask and add dimensions
        question_mask, reply_input_mask, reply_target_mask = create_masks(question, reply_input, reply_target)

        # Get the transformer outputs
        out = transformer(question, question_mask, reply_input, reply_input_mask)

        # Compute the loss
        loss = criterion(out, reply_target, reply_target_mask)
        
        # Backprop
        transformer_optimizer.optimizer.zero_grad()
        loss.backward()
        transformer_optimizer.step()
        
        sum_loss += loss.item() * samples
        count += samples
        
        if i % 100 == 0:
            print("Epoch [{}][{}/{}]\tLoss: {:.3f}".format(epoch, i, len(train_loader), sum_loss/count))
            
            
for epoch in range(epochs):
    
    train(train_loader, transformer, criterion, epoch)
    
    state = {'epoch': epoch, 'transformer': transformer, 'transformer_optimizer': transformer_optimizer}
    torch.save(state, 'checkpoint_' + str(epoch) + '.pth.tar')



import random
import csv
import string
def preprocess(text):
    # Convert text to lowercase and split into words
    words = text.lower().split()

    # Create dictionary mapping each unique word to a unique integer index
    word_to_index = {word: i for i, word in enumerate(set(words))}

    # Convert each word in the text to its corresponding integer index
    indices = [word_to_index[word] for word in words]

    # Convert integer indices to one-hot vectors
    num_words = len(word_to_index)
    input_vectors = np.zeros((len(indices), num_words))
    input_vectors[np.arange(len(indices)), indices] = 1

    return input_vectors, word_to_index

def preprocessorx(list):
    import random

    my_list = list

    used_numbers = []

    for i in range(len(my_list)):
        if my_list[i] in used_numbers:
            continue
        else:
            random_number = random.randint(1, 10000)
            while random_number in used_numbers:
                random_number = random.randint(1, 10000)
            my_list[i] = str(random_number)
            used_numbers.append(random_number)

    print(my_list)


import json

def extractor(text):
    text = text.split("\n")
    text = [s.split(" ") for s in text]
    words = []
    for sentence in text:
        for word in sentence:
            words.append(word)
    return words

def preprocessor(words):
    used_values = {}
    output = {}
    for word in words:
        if word in used_values:
            output[word] = used_values[word]
        else:
            value = round(random.uniform(0, 1), 8)
            used_values[word] = value
            output[word] = value

    with open('preprocessed_fr_eng.json', 'w') as f:
        json.dump(output, f)

#with open('input.txt', 'r') as f:
    #text = f.read()

#words = extractor(text)
#preprocessor(words)
import numpy as np

# Sigmoid activation function
def sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))

# Derivative of sigmoid function
def sigmoid_derivative(x):
    return sigmoid(x) * (1.0 - sigmoid(x))

class NeuralNetwork:
    def __init__(self, input_size, hidden_size, output_size):
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.output_size = output_size

        # Initialize weights
        self.weights1 = np.random.rand(self.input_size, self.hidden_size)
        self.weights2 = np.random.rand(self.hidden_size, self.output_size)

    def feedforward(self, X):
        # Propagate input through the network
        self.hidden_layer = sigmoid(np.dot(X, self.weights1))
        self.output_layer = sigmoid(np.dot(self.hidden_layer, self.weights2))
        return self.output_layer

    def backpropagation(self, X, y, learning_rate):
        # Calculate errors and deltas for each layer
        output_error = y - self.output_layer
        output_delta = output_error * sigmoid_derivative(self.output_layer)

        hidden_error = np.dot(output_delta, self.weights2.T)
        hidden_delta = hidden_error * sigmoid_derivative(self.hidden_layer)

        # Update weights with deltas and learning rate
        self.weights2 += learning_rate * np.dot(self.hidden_layer.T, output_delta)
        self.weights1 += learning_rate * np.dot(X.T, hidden_delta)

    def train(self, X, y, learning_rate, epochs):
        for i in range(epochs):
            # Feedforward through the network

            output = self.feedforward(X)

            # Backpropagation to update weights
            self.backpropagation(X, y, learning_rate)
            print("Iterating Epoch :", i,"/",epochs)

    def predict(self, X):
        # Predict output for new input
        return self.feedforward(X)





# Define neural network class
class NeuralNetwork2:
    def __init__(self, input_size, hidden_size, output_size):
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.output_size = output_size

        # Initialize weights
        self.weights1 = np.random.rand(self.input_size, self.hidden_size)
        self.weights2 = np.random.rand(self.hidden_size, self.output_size)

    def feedforward(self, X):
        # Propagate input through the network
        self.hidden_layer = sigmoid(np.dot(X, self.weights1))
        self.output_layer = sigmoid(np.dot(self.hidden_layer, self.weights2))
        return self.output_layer

    def backpropagation(self, X, y, learning_rate):
        # Calculate errors and deltas for each layer
        output_error = y - self.output_layer
        output_delta = output_error * sigmoid_derivative(self.output_layer)

        hidden_error = np.dot(output_delta, self.weights2.T)
        hidden_delta = hidden_error * sigmoid_derivative(self.hidden_layer)

        # Update weights with deltas and learning rate
        self.weights2 += learning_rate * np.dot(self.hidden_layer.T, output_delta)
        self.weights1 += learning_rate * np.dot(X.T, hidden_delta)

    def train(self, X, y, learning_rate, epochs):
        for i in range(epochs):
            # Feedforward through the network
            output = self.feedforward(X)

            # Backpropagation to update weights
            self.backpropagation(X, y, learning_rate)
            print("Iterating Epoch: ", i, " / ", epochs)

    def predict(self, X):
        # Predict output for new input
        return self.feedforward(X)




def rxgenv_model(corpus,seed,len1=10,len2=20,learning_rate=0.1,hidden_size=100,epochs=10,save="save.npz",traintype="cpu",pretrain="False",pretrain_file="save.npz"):
    corpus = open(corpus).read().lower().translate(str.maketrans('', '', string.punctuation)).split()
    if traintype=="cpu":
        xp = np
    elif traintype=="gpu":
        import cupy as cp
        xp = cp
    elif traintype=="CPU":
        xp = np
    elif traintype=="GPU":
        import cupy as cp
        xp = cp
    else:
        xp=np

    vocab = sorted(set(corpus))
    word2idx = {w: i for i, w in enumerate(vocab)}
    idx2word = {i: w for i, w in enumerate(vocab)}


    WINDOW_SIZE = 2
    train_data = []
    for i in range(len(corpus) - WINDOW_SIZE):
        train_data.append(([corpus[j] for j in range(i, i + WINDOW_SIZE)], corpus[i + WINDOW_SIZE]))


    INPUT_SIZE = len(vocab)
    HIDDEN_SIZE = hidden_size
    OUTPUT_SIZE = len(vocab)
    LEARNING_RATE = learning_rate

    if pretrain == "True":
        data = xp.load(pretrain_file)
        U = data['U']
        V = data['V']
        W = data['W']
        bh = data['bh']
        by = data['by']

    elif pretrain == "False":
        U = xp.random.randn(INPUT_SIZE, HIDDEN_SIZE) * 0.01
        V = xp.random.randn(HIDDEN_SIZE, OUTPUT_SIZE) * 0.01
        W = xp.random.randn(HIDDEN_SIZE, HIDDEN_SIZE) * 0.01
        bh = xp.zeros((1, HIDDEN_SIZE))
        by = xp.zeros((1, OUTPUT_SIZE))
    else:
        U = xp.random.randn(INPUT_SIZE, HIDDEN_SIZE) * 0.01
        V = xp.random.randn(HIDDEN_SIZE, OUTPUT_SIZE) * 0.01
        W = xp.random.randn(HIDDEN_SIZE, HIDDEN_SIZE) * 0.01
        bh = xp.zeros((1, HIDDEN_SIZE))
        by = xp.zeros((1, OUTPUT_SIZE))


    for epoch in range(epochs):
        loss = 0
        for X, y in train_data:

            x_idx = [word2idx[w] for w in X]
            y_idx = word2idx[y]


            h = xp.zeros((1, HIDDEN_SIZE))
            inputs = xp.zeros((1, INPUT_SIZE))
            for i in x_idx:
                inputs[0, i] = 1


            h = xp.tanh(xp.dot(inputs, U) + xp.dot(h, W) + bh)
            y_hat = xp.exp(xp.dot(h, V) + by) / xp.sum(xp.exp(xp.dot(h, V) + by))


            loss += -xp.log(y_hat[0, y_idx])


            dy = xp.copy(y_hat)
            dy[0, y_idx] -= 1
            dV = xp.dot(h.T, dy)
            dby = dy
            dh = xp.dot(dy, V.T)
            dg = (1 - h ** 2) * dh
            dU = xp.dot(inputs.T, dg)
            dbh = dg


            U -= LEARNING_RATE * dU
            V -= LEARNING_RATE * dV
            W -= LEARNING_RATE * dh
            bh -= LEARNING_RATE * dbh
            by -= LEARNING_RATE * dby
        correct_count = 0
        total_count = 0
        for X, y in train_data:
            x_idx = [word2idx[w] for w in X]
            y_idx = word2idx[y]
            h = xp.zeros((1, HIDDEN_SIZE))
            inputs = xp.zeros((1, INPUT_SIZE))
            for i in x_idx:
                inputs[0, i] = 1
            h = xp.tanh(xp.dot(inputs, U) + xp.dot(h, W) + bh)
            y_hat = xp.exp(xp.dot(h, V) + by) / xp.sum(xp.exp(xp.dot(h, V) + by))
            predicted_idx = xp.argmax(y_hat)
            if predicted_idx == y_idx:
                correct_count += 1
            total_count += 1
        accuracy = correct_count / total_count
        print('Epoch:', epoch+ 1, '/ ', epochs, ' \n Loss:', loss, 'Accuracy:', accuracy )
    xp.savez(save, U=U, V=V, W=W, bh=bh, by=by, word2idx=word2idx, idx2word=idx2word)
    for i in range(len1):
        x = xp.zeros((1, INPUT_SIZE))
        x[0, word2idx[seed]] = 1
        h = xp.zeros((1, HIDDEN_SIZE))
        sentence = [seed]
        for j in range(len2):
            h = xp.tanh(xp.dot(x, U) + xp.dot(h, W) + bh)
            y = xp.dot(h, V) + by
            p = xp.exp(y) / xp.sum(xp.exp(y))
            idx = xp.random.choice(range(len(vocab)), p=p.ravel())
            if idx2word[idx] == '.':
                sentence.append('.')
                break
            sentence.append(idx2word[idx])
            x = xp.zeros((1, INPUT_SIZE))
            x[0, idx] = 1

        return  ' '.join(sentence)
#print(model("internet_archive_scifi_v3.txt","the",epochs=100,hidden_size=50,learning_rate=0.1))





from collections import Counter
import json
import csv

max_len = 25

def remove_punc(string):
    punctuations = '''!()-[]{};:'"\,<>./?@#$%^&*_~'''
    no_punct = ""
    for char in string:
        if char not in punctuations:
            no_punct = no_punct + char  # space is also a character
    return no_punct.lower()


pairs = []
# Open the CSV file for reading
with open('file.csv', 'r' ,errors='ignore') as csv_file:
    # Create a CSV reader object
    csv_reader = csv.reader(csv_file)

    # Get the header row
    header = next(csv_reader)

    # Find the index of the 'input' and 'label' columns
    input_index = header.index('input')
    label_index = header.index('label')

    # Loop through each row in the CSV file
    for row in csv_reader:
        qa_pairs = []
        # Extract the text from the 'input' and 'label' columns at the same index
        input_text = row[input_index]
        label_text = row[label_index]
        first = remove_punc(input_text.strip())
        second = remove_punc(label_text.strip())
        qa_pairs.append(first.split()[:max_len])
        qa_pairs.append(second.split()[:max_len])
        pairs.append(qa_pairs)
        # Do something with the text

# print(pairs)
word_freq = Counter()
for pair in pairs:
    word_freq.update(pair[0])
    word_freq.update(pair[1])

min_word_freq = 1
words = [w for w in word_freq.keys() if word_freq[w] > min_word_freq]
word_map = {k: v + 1 for v, k in enumerate(words)}
word_map['<unk>'] = len(word_map) + 1
word_map['<start>'] = len(word_map) + 1
word_map['<end>'] = len(word_map) + 1
word_map['<pad>'] = 0

print("Total words are: {}".format(len(word_map)))

with open('WORDMAP_corpus.json', 'w') as j:
    json.dump(word_map, j)


def encode_question(words, word_map):
    enc_c = [word_map.get(word, word_map['<unk>']) for word in words] + [word_map['<pad>']] * (max_len - len(words))
    return enc_c


def encode_reply(words, word_map):
    enc_c = [word_map['<start>']] + [word_map.get(word, word_map['<unk>']) for word in words] + [word_map['<end>']] + [
        word_map['<pad>']] * (max_len - len(words))
    return enc_c


pairs_encoded = []
for pair in pairs:
    qus = encode_question(pair[0], word_map)
    ans = encode_reply(pair[1], word_map)
    pairs_encoded.append([qus, ans])

with open('pairs_encoded.json', 'w') as p:
    json.dump(pairs_encoded, p)



import torch
import torch.nn as nn
from torch.utils.data import Dataset
import torch.utils.data
import json

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class Dataset(Dataset):

    def __init__(self):

        self.pairs = json.load(open('pairs_encoded.json'))
        self.dataset_size = len(self.pairs)

    def __getitem__(self, i):
        
        question = torch.LongTensor(self.pairs[i][0])
        reply = torch.LongTensor(self.pairs[i][1])
            
        return question, reply

    def __len__(self):
        return self.dataset_size
    

def create_masks(question, reply_input, reply_target):
    
    def subsequent_mask(size):
        mask = torch.triu(torch.ones(size, size)).transpose(0, 1).type(dtype=torch.uint8)
        return mask.unsqueeze(0)
    
    question_mask = (question!=0).to(device)
    question_mask = question_mask.unsqueeze(1).unsqueeze(1)         # (batch_size, 1, 1, max_words)
     
    reply_input_mask = reply_input!=0
    reply_input_mask = reply_input_mask.unsqueeze(1)  # (batch_size, 1, max_words)
    reply_input_mask = reply_input_mask & subsequent_mask(reply_input.size(-1)).type_as(reply_input_mask.data) 
    reply_input_mask = reply_input_mask.unsqueeze(1) # (batch_size, 1, max_words, max_words)
    reply_target_mask = reply_target!=0              # (batch_size, max_words)
    
    return question_mask, reply_input_mask, reply_target_mask


class AdamWarmup:
    
    def __init__(self, model_size, warmup_steps, optimizer):
        
        self.model_size = model_size
        self.warmup_steps = warmup_steps
        self.optimizer = optimizer
        self.current_step = 0
        self.lr = 0
        
    def get_lr(self):
        return self.model_size ** (-0.5) * min(self.current_step ** (-0.5), self.current_step * self.warmup_steps ** (-1.5))
        
    def step(self):
        # Increment the number of steps each time we call the step function
        self.current_step += 1
        lr = self.get_lr()
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr
        # update the learning rate
        self.lr = lr
        self.optimizer.step()   

class LossWithLS(nn.Module):

    def __init__(self, size, smooth):
        super(LossWithLS, self).__init__()
        self.criterion = nn.KLDivLoss(size_average=False, reduce=False)
        self.confidence = 1.0 - smooth
        self.smooth = smooth
        self.size = size
        
    def forward(self, prediction, target, mask):
        """
        prediction of shape: (batch_size, max_words, vocab_size)
        target and mask of shape: (batch_size, max_words)
        """
        prediction = prediction.view(-1, prediction.size(-1))   # (batch_size * max_words, vocab_size)
        target = target.contiguous().view(-1)   # (batch_size * max_words)
        mask = mask.float()
        mask = mask.view(-1)       # (batch_size * max_words)
        labels = prediction.data.clone()
        labels.fill_(self.smooth / (self.size - 1))
        labels.scatter_(1, target.data.unsqueeze(1), self.confidence)
        loss = self.criterion(prediction, labels)    # (batch_size * max_words, vocab_size)
        loss = (loss.sum(1) * mask).sum() / mask.sum()
        return loss



# Library by Webraft-AI
import nltk
from nltk.tokenize import word_tokenize, sent_tokenize
from random import choice
from difflib import SequenceMatcher

def generator(user_input,threshold):


    # Define initial lists of words
    noun_phrases = ["the cat", "the dog", "the bird", "the tree"]
    verbs = ["is", "looks", "seems", "goes", "works", "hacks", "lags", "watches"]
    adjectives = ["happy", "sad", "angry", "excited", "busy", "creative", "beautiful","glad"]
    adverbs = ["very", "quite", "extremely", "much", "very much", "confidently", "gracefully"]

    # Define a similarity threshold for generated sentences
    similarity_threshold = threshold

    # Get input from user and split it into words

    user_words = word_tokenize(user_input.lower())

    # Append user words to appropriate lists
    for word in user_words:
        if word in noun_phrases:
            continue
        elif word in verbs:
            continue
        elif word in adjectives:
            continue
        elif word in adverbs:
            continue
        else:
            pos_tag = nltk.pos_tag([word])[0][1]
            if pos_tag.startswith("NN"):
                noun_phrases.append(word)
            elif pos_tag.startswith("VB"):
                verbs.append(word)
            elif pos_tag.startswith("JJ"):
                adjectives.append(word)
            elif pos_tag.startswith("RB"):
                adverbs.append(word)

    # Generate sentences until one is similar enough to the user input
    generated_sentence = ""
    while True:
        sentence_structure = choice([1, 2, 3])
        if sentence_structure == 1:
            np = choice(noun_phrases)
            v = choice(verbs)
            adj = choice(adjectives)
            generated_sentence = f"{np} {v} {adj}."
            pos_tags = nltk.pos_tag(word_tokenize(generated_sentence))
            if pos_tags[-1][1].startswith("NN"):
                noun_phrases.append(pos_tags[-1][0])
                generated_sentence = f"{np} {v} {adj} {pos_tags[-1][0]}."
        elif sentence_structure == 2:
            np1 = choice(noun_phrases)
            np2 = choice(noun_phrases)
            v = choice(verbs)
            generated_sentence = f"{np1} {v} like {np2}."
            pos_tags = nltk.pos_tag(word_tokenize(generated_sentence))
            if pos_tags[-1][1].startswith("NN"):
                noun_phrases.append(pos_tags[-1][0])
                generated_sentence = f"{np1} {v} like {np2} {pos_tags[-1][0]}."
        else:
            adj = choice(adjectives)
            adv = choice(adverbs)
            v = choice(verbs)
            generated_sentence = f"It {v} {adv} {adj}."
            pos_tags = nltk.pos_tag(word_tokenize(generated_sentence))
            if pos_tags[-1][1].startswith("NN"):
                noun_phrases.append(pos_tags[-1][0])

        # Calculate similarity between generated sentence and user input
        similarity_ratio = SequenceMatcher(None, user_input, generated_sentence).ratio()

        if similarity_ratio >= similarity_threshold:
            return generated_sentence
def autogen(number):

    from random import choice

    # Define lists of words for each sentence structure
    noun_phrases = ["the cat", "the dog", "the bird", "the tree"]
    verbs = ["is", "looks", "seems"]
    adjectives = ["happy", "sad", "angry", "excited"]
    adverbs = ["very", "quite", "extremely"]

    # Generate a sentence using a random sentence structure
    def generate_sentence():
        sentence_structure = choice([1, 2, 3])
        if sentence_structure == 1:
            np = choice(noun_phrases)
            v = choice(verbs)
            adj = choice(adjectives)
            sentence = f"{np} {v} {adj}."
        elif sentence_structure == 2:
            np1 = choice(noun_phrases)
            np2 = choice(noun_phrases)
            v = choice(verbs)
            sentence = f"{np1} {v} like {np2}."
        else:
            adj = choice(adjectives)
            adv = choice(adverbs)
            v = choice(verbs)
            sentence = f"It {v} {adv} {adj}."
        return sentence

    # Generate and print 10 sentences
    for i in range(number):
        return generate_sentence()

def customgen(nouns,verbs,adjectives,adverbs,number=1):
    from random import choice

    # Define lists of words for each sentence structure
    noun_phrases = nouns


    # Generate a sentence using a random sentence structure
    def generate_sentence():
        sentence_structure = choice([1, 2, 3])
        if sentence_structure == 1:
            np = choice(noun_phrases)
            v = choice(verbs)
            adj = choice(adjectives)
            sentence = f"{np} {v} {adj}."
        elif sentence_structure == 2:
            np1 = choice(noun_phrases)
            np2 = choice(noun_phrases)
            v = choice(verbs)
            sentence = f"{np1} {v} like {np2}."
        else:
            adj = choice(adjectives)
            adv = choice(adverbs)
            v = choice(verbs)
            sentence = f"It {v} {adv} {adj}."
        return sentence

    # Generate and print 10 sentences
    for i in range(number):
        return generate_sentence()

def custgenx(nouns,verbs,adjectives,adverbs,user_input,threshold=0.7):


    # Define initial lists of words
    noun_phrases = nouns

    # Define a similarity threshold for generated sentences
    similarity_threshold = threshold

    # Get input from user and split it into words

    user_words = word_tokenize(user_input.lower())

    # Append user words to appropriate lists
    for word in user_words:
        if word in noun_phrases:
            continue
        elif word in verbs:
            continue
        elif word in adjectives:
            continue
        elif word in adverbs:
            continue
        else:
            pos_tag = nltk.pos_tag([word])[0][1]
            if pos_tag.startswith("NN"):
                noun_phrases.append(word)
            elif pos_tag.startswith("VB"):
                verbs.append(word)
            elif pos_tag.startswith("JJ"):
                adjectives.append(word)
            elif pos_tag.startswith("RB"):
                adverbs.append(word)

    # Generate sentences until one is similar enough to the user input
    generated_sentence = ""
    while True:
        sentence_structure = choice([1, 2, 3])
        if sentence_structure == 1:
            np = choice(noun_phrases)
            v = choice(verbs)
            adj = choice(adjectives)
            generated_sentence = f"{np} {v} {adj}."
            pos_tags = nltk.pos_tag(word_tokenize(generated_sentence))
            if pos_tags[-1][1].startswith("NN"):
                noun_phrases.append(pos_tags[-1][0])
                generated_sentence = f"{np} {v} {adj} {pos_tags[-1][0]}."
        elif sentence_structure == 2:
            np1 = choice(noun_phrases)
            np2 = choice(noun_phrases)
            v = choice(verbs)
            generated_sentence = f"{np1} {v} like {np2}."
            pos_tags = nltk.pos_tag(word_tokenize(generated_sentence))
            if pos_tags[-1][1].startswith("NN"):
                noun_phrases.append(pos_tags[-1][0])
                generated_sentence = f"{np1} {v} like {np2} {pos_tags[-1][0]}."
        else:
            adj = choice(adjectives)
            adv = choice(adverbs)
            v = choice(verbs)
            generated_sentence = f"It {v} {adv} {adj}."
            pos_tags = nltk.pos_tag(word_tokenize(generated_sentence))
            if pos_tags[-1][1].startswith("NN"):
                noun_phrases.append(pos_tags[-1][0])

        # Calculate similarity between generated sentence and user input
        similarity_ratio = SequenceMatcher(None, user_input, generated_sentence).ratio()

        if similarity_ratio >= similarity_threshold:
            return generated_sentence


import json
import torch
from torch.utils.data import Dataset
import torch.utils.data


load_checkpoint = True
ckpt_path = 'checkpoint.pth.tar'
with open('WORDMAP_corpus.json', 'r') as j:
    word_map = json.load(j)

def evaluate(transformer, question, question_mask, max_len, word_map):
    """
    Performs Greedy Decoding with a batch size of 1
    """
    rev_word_map = {v: k for k, v in word_map.items()}
    transformer.eval()
    start_token = word_map['<start>']
    encoded = transformer.encode(question, question_mask)
    words = torch.LongTensor([[start_token]]).to(device)
    
    for step in range(max_len - 1):
        size = words.shape[1]
        target_mask = torch.triu(torch.ones(size, size)).transpose(0, 1).type(dtype=torch.uint8)
        target_mask = target_mask.to(device).unsqueeze(0).unsqueeze(0)
        decoded = transformer.decode(words, target_mask, encoded, question_mask)
        predictions = transformer.logit(decoded[:, -1])
        _, next_word = torch.max(predictions, dim = 1)
        next_word = next_word.item()
        if next_word == word_map['<end>']:
            break
        words = torch.cat([words, torch.LongTensor([[next_word]]).to(device)], dim = 1)   # (1,step+2)
        
    # Construct Sentence
    if words.dim() == 2:
        words = words.squeeze(0)
        words = words.tolist()
        
    sen_idx = [w for w in words if w not in {word_map['<start>']}]
    sentence = ' '.join([rev_word_map[sen_idx[k]] for k in range(len(sen_idx))])
    
    return sentence


if load_checkpoint:
    checkpoint = torch.load(ckpt_path)
    transformer = checkpoint['transformer']


while(1):
    question = input("Question: ") 
    if question == 'quit':
        break
    max_len = input("Maximum Reply Length: ")
    enc_qus = [word_map.get(word, word_map['<unk>']) for word in question.split()]
    question = torch.LongTensor(enc_qus).to(device).unsqueeze(0)
    question_mask = (question!=0).to(device).unsqueeze(1).unsqueeze(1)  
    sentence = evaluate(transformer, question, question_mask, int(max_len), word_map)
    print(sentence)

